
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import org.apache.xml.serialize.OutputFormat;
import org.apache.xml.serialize.XMLSerializer;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import java.io.*;
import java.net.URL;
import jdrew.oo.td.*;
import jdrew.oo.util.*;
import java.util.*;
import org.apache.commons.httpclient.*;
import org.apache.commons.httpclient.methods.*;
import resources_PA.*;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.*;
import com.oreilly.servlet.HttpMessage;

/**
 *
 * Wrapper agent for the DR-DEVICE defeasible reasoner.
 * This agent serves as a PA in the Symposium Planner RR application.
 *
 * @author Efstratios Kontopoulos <skontopo@csd.auth.gr>
 */
public class DRDChairRuleML2010 extends HttpServlet {

    // PATH & PROJECT PARAMETERS
    private String queryPath = "webapps\\drdfiles\\queries\\";
    private String xsltFileName = "webapps\\drdfiles\\core\\dr-device-input.xslt";
    private String exportPath = "webapps\\drdfiles\\results\\";
    private String defaultPath = "C:/rrdemo/";
    private String ruleBase = "sponsors.ruleml";
    private String export = "sponsors_output.rdf";
    private String projectName;
    private Document doc;

    // DR-DEVICE GLOBAL PARAMETERS - do not modify
    private String system = "dr-device";
    private String verbose = "(set-verbose on)";
    private String debug = "(set-debug off)";
    private String time_report = "(set-time-report off)";
    //private String compact_proofs = "(set-compact-proofs on)";
    //private String export_non_proved = "(set-export-non-proved off)";
    private String namespace_hunting = "(set-namespace-hunting off)";
    public boolean local = true;

    // PA PARAMETERS
    private final String address = "http://155.207.113.24"; // SymposiumPlanner-2010
    private final String port = "9998"; // SymposiumPlanner-2010
    public final static String FS = System.getProperty("file.separator");
    private final String poslAddress = "http://155.207.113.24:8080/ruleSetsRuleML-2010/generalChairRuleML-2010.posl"; // SymposiumPlanner-2010
    private final String rdfAddress = "http://155.207.113.24:8080/ruleSetsRuleML-2010/publicityChairOntologyRuleML-2010.rdf"; // SymposiumPlanner-2010

    public void init(ServletConfig config) throws ServletException {
        //super.init(config);
    }

    public void doGet(HttpServletRequest request, HttpServletResponse response)
            throws ServletException {
        response.setContentType("text/html; charset=UTF-8");

        try {
            PrintWriter out = response.getWriter();
            out.println("DR-DEVICE Servlet RuleML-2010");
        } catch (Exception ex) {
            System.out.println(ex.getMessage());
        }
    }

    public void doPost(HttpServletRequest request, HttpServletResponse response)
            throws ServletException {
        response.setContentType("text/html; charset=UTF-8");

        try {
            BufferedReader brd = request.getReader();

            String input = "";
            String message = "";

            while (!input.equals("</RuleML>")) {
                input = brd.readLine();
                message = message + input;
            }
            System.out.println("Received Message:\n" + message);

            // Each new project is assigned a different (random) name
            projectName = "project" + "-" + randomString();
            System.out.println("New DR-DEVICE project created: " + projectName);

            // Creates a file containing the query.
            // This file will then be transformed via XSLT into an RDF input
            // document for the DR-DEVICE reasoner.
            createFile(message, queryPath, "query-" + projectName);

            // XSL Transformation: Creates input RDF document
            xslTransformation(queryPath + "query-" + projectName);

            // Create project workspace
            makeDir();
            createRbFile();
            writeToFile();

            // Launch DR-DEVICE reasoner
            runBat(projectName);

            // Retrieve inference results.
            String exportRDF = rdfFileContent();
            System.out.println("Output:\n" + exportRDF);
            String path = createFile(exportRDF, exportPath, "results-" + projectName + ".rdf");

            System.out.println("DRD-Agent: Sending message");
            sendMessage(message, path);

        } catch (Exception ex) {
            System.out.println(ex.getMessage());
        }
    }

    private void sendMessage(String message, String filePath) throws Exception {
        String[] varOrder = getVariableOrder(message);

        POSLParser pp = new POSLParser();

        Date t1 = new GregorianCalendar().getTime();

        //time
        String time = "time(" + t1.getHours() + ":integer).";
        String contents = "";

        //day of the week
        int day = t1.getDay();
        boolean weekday = true;

        if (day == 0 || day == 6) {
            weekday = false;
        }

        String dayOfWeek;

        if (weekday) {
            dayOfWeek = "day(weekday).";
        } else {
            dayOfWeek = "day(weekend).";
        }
        //full date
        Calendar cal = new GregorianCalendar();

        int year = cal.get(Calendar.YEAR);
        int month = cal.get(Calendar.MONTH) + 1;
        int day2 = cal.get(Calendar.DAY_OF_MONTH);

        String date;

        String day3 = "" + day2;

        if (day2 == 1 || day2 == 2 || day2 == 3 ||
                day2 == 4 || day2 == 5 || day2 == 6 ||
                day2 == 7 || day2 == 8 || day2 == 9) {

            day3 = "0" + day2;
        }

        if (month == 10 || month == 11 || month == 12) {
            date = "" + year + month + day3;
        } else {
            date = "" + year + "0" + month + day3;
        }

        date = "date(" + date + ":integer).";

        String url = poslAddress;
        String url2 = rdfAddress;
        HttpClient client2 = new HttpClient();
        GetMethod method2 = new GetMethod(url2);
        method2.setFollowRedirects(true);
        String typestr = "";

        // Execute the GET method
        int statusCode2 = client2.executeMethod(method2);
        if (statusCode2 != -1) {
            typestr = method2.getResponseBodyAsString();
        }
        Types.reset();
        RDFSParser.parseRDFSString(typestr);

        try {
            HttpClient client = new HttpClient();
            GetMethod method = new GetMethod(url);
            method.setFollowRedirects(true);

            // Execute the GET method
            int statusCode = client.executeMethod(method);
            if (statusCode != -1) {
                contents = method.getResponseBodyAsString();
            }
        } catch (Exception e) {
            e.printStackTrace();
        }

        contents = contents + "\n" + time;
        contents = contents + "\n" + dayOfWeek;
        contents = contents + "\n" + date;

        BackwardReasoner br = new BackwardReasoner();

        SymbolTable.reset();

        pp.parseDefiniteClauses(contents);
        br.loadClauses(pp.iterator());

        Iterator it = pp.iterator();
        while (it.hasNext()) {
            DefiniteClause d = (DefiniteClause) it.next();
        }

        br = new BackwardReasoner(br.clauses, br.oids);

        MessageParser m = new MessageParser(message);
        nu.xom.Element atom = null;

        try {
            atom = m.parseForContent();
        } catch (Exception e) {
            System.out.println("Invalid Message");
        //out.flush();
        }

        QueryBuilder q = new QueryBuilder(atom);
        String query = q.generateDoc();
        //System.out.println("ABOUT TO INPUT THIS QUERY:" + query);
        RuleMLParser qp = new RuleMLParser();

        DefiniteClause dc = null;
        try {
            dc = qp.parseRuleMLQuery(query);
        } catch (Exception e) {
            System.out.println("Invalid Query");
        //out.flush();
        }

        int varSize = 0;

        MessageGenerator g2 = new MessageGenerator(
                null, varSize, "ruleml2010_DRDChair",
                m.getId(), m.getProtocol(), m.getRel());

        URL sender2 = new URL(address + ":" + port);
        HttpMessage msg2 = new HttpMessage(sender2);
        Properties props2 = new Properties();

        String finalMessage2 = g2.finalMessage2(query, filePath);

        props2.put("text", finalMessage2);
        InputStream in2 = msg2.sendGetMessage(props2);

        MessageGenerator g = new MessageGenerator(
                null, varSize, "ruleml2010_DRDChair",
                m.getId(), m.getProtocol(), m.getRel());

        URL sender = new URL(address + ":" + port);
        HttpMessage msg = new HttpMessage(sender);
        Properties props = new Properties();

        String finalMessage = g.finalMessage(query);

        props.put("text", finalMessage);
        InputStream in = msg.sendGetMessage(props);
    }

    // Creates a file containing an input string.
    private String createFile(String fileContent, String path, String fileName)
            throws Exception {
        File file = new File(path + fileName);
        FileWriter out = new FileWriter(file);
        out.write(fileContent);
        out.close();
        return file.getCanonicalPath();
    }

    // Returns a random 4-character string:
    // Characters 1-2 are letters, characters 3-4 are numbers.
    private String randomString() {
        java.util.Random random = new java.util.Random();
        String rstr = "";
        for (int j = 0; j < 2; j++) {
            int rand = random.nextInt(25);
            rstr += (char) (97 + rand);
        }
        for (int j = 0; j < 2; j++) {
            int rand = random.nextInt(10);
            rstr += rand;
        }
        return rstr;
    }

    // XSL Transformation.
    private void xslTransformation(String xml) {
        try {
            transform(xml, xsltFileName);
        } catch (Exception ex) {
            System.out.println("xslTransformation Exception:");
            System.out.println(ex.getMessage());
            return;
        }
        System.out.println("Message transformed.");
    }

    // Auxilliary method for the xslTransformation(String xml) method.
    private void transform(String xml, String xslt) throws Exception {

        // 1. Instantiate a TransformerFactory.
        javax.xml.transform.TransformerFactory tFactory =
                javax.xml.transform.TransformerFactory.newInstance();

        // 2. Use the TransformerFactory to process the stylesheet Source and
        //    generate a Transformer.
        javax.xml.transform.Transformer transformer =
                tFactory.newTransformer(
                new javax.xml.transform.stream.StreamSource(xslt));

        // 3. Use the Transformer to transform an XML Source and send the
        //    output to a Result object.
        transformer.transform(new javax.xml.transform.stream.StreamSource(xml),
                new javax.xml.transform.stream.StreamResult(
                new java.io.FileOutputStream(
                defaultPath + "sponsors-input-" + projectName + ".rdf")));
    }

    // Create the directory, where the specific project files will be stored.
    private void makeDir() {
        File file = new File("C:\\Program Files\\DR-Device\\projects/" + projectName);
        try {
            file.mkdirs();
        } catch (Exception e) {
            System.out.println("Could not create project directory.");
            System.out.println(e.getMessage());
            return;
        }
        System.out.println("Folder " + projectName + " created.");
    }

    // Creates a copy of the input rule base inside the project workspace.
    private void createRbFile() {
        try {
            DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
            DocumentBuilder builder = factory.newDocumentBuilder();
            doc = builder.parse(new File(defaultPath + "" + ruleBase));
            Element root = doc.getDocumentElement();
            root.setAttribute(
                    "rdf_import",
                    "file:///" + defaultPath + "sponsors-input-" + projectName + ".rdf");

            File file = new File("C:\\Program Files\\DR-Device\\projects/" + projectName + "/" + projectName + ".ruleml");
            file.delete(); // delete the file if it already exists
            FileWriter out = new FileWriter(file);
            OutputFormat oformat = new OutputFormat(doc);
            oformat.setIndenting(true);
            oformat.setIndent(2);
            oformat.setLineWidth(65);
            XMLSerializer serializer = new XMLSerializer(out, oformat);
            serializer.serialize(doc);
            out.flush();
            out.close();

        } catch (Exception ex) {
            System.out.println("Cannot create rule base file");
            System.out.println(ex.getMessage());
            ex.printStackTrace();
            return;
        }
        System.out.println("Rule base created.");
    }

    // Creates the first .bat files in the relevant directory.
    private void writeToFile() {
        PrintWriter pw = null;
        PrintWriter pw2 = null;
        String url = defaultPath + "" + ruleBase;

        String dir = "(system \"cmd /c echo %dr-device% > current-dir\")\n" +
                "(open \"current-dir\" file \"r\")\n" +
                "(progn\n" +
                "(bind ?dir (readline file))\n" +
                "(open \"run-" + system + ".bat\" run \"w\")\n" +
                "(printout run (str-cat \"(batch* \\\"\" ?dir \"\\\\bin\\\\" + system + ".bat\\\")\") crlf)\n" +
                "(close run)\n" +
                ")\n" +
                "(close file)\n" +
                "(system \"cmd /c \\\"\\\"%dr-device%\\\\bin\\\\repair-path.exe\\\" run-" + system + ".bat\\\"\")\n" +
                "(dribble-on \"" + projectName + ".log\")\n" +
                "(batch* \"run-" + system + ".bat\")\n" +
                "(remove \"current-dir\")\n";

        File file = new File("C:\\Program Files\\DR-Device\\projects/" + projectName + "/" + projectName + ".bat");
        File file2 = new File("C:\\Program Files\\DR-Device\\projects/" + projectName + "/" + "run-" + projectName + ".bat");

        // Delete the old instances of the files
        file.delete();
        file2.delete();

        try {
            pw = new PrintWriter(new FileOutputStream(file, true));
            pw.print(dir);
            if (!this.system.equals("dr-device")) {
                pw.print(verbose + "\n" + debug + "\n" + time_report + "\n" +
                        namespace_hunting + "\n");
            } else {
                pw.print(verbose + "\n" + debug + "\n" + time_report + "\n" +
                        //compact_proofs + "\n" + export_non_proved + "\n" +
                        namespace_hunting + "\n");
            }

            if (local == false) {
                pw.print("(load-ruleml-" + system + " \"" + projectName + "\" \"" +
                        url + "\")" + "\n" + "(dribble-off)\n" + "(exit)" + "\n");
            } else {
                pw.print("(load-ruleml-" + system + "-local \"" + projectName + "\" \"" +
                        transform(url) + "\")" + "\n" + "(dribble-off)\n" + "(exit)" + "\n");
            }
            pw2 = new PrintWriter(new FileOutputStream(file2, true));
            pw2.print("@cd %dr-device%\\projects\\" + projectName + "\n");
            pw2.print("@\"%dr-device%\\CLIPSDOS\\CLIPSDOS.exe\" -f2 .\\" + projectName + ".bat\n");
            pw2.print("@pause\nexit\n");
        } catch (IOException ex) {
            System.out.println("An error occured while writing files.");
            System.out.println(ex.getMessage());
            return;
        } finally {
            if (pw != null) {
                pw.close();
            }
            if (pw2 != null) {
                pw2.close();
            }
        }
        System.out.println(".bat files created.");
    }

    // Converts the "\" character into "\\".
    private static String transform(String s) {
        StringBuffer sb = new StringBuffer(s);
        String result = "";

        for (int i = 0; i < s.length(); i++) {
            Character c = new Character(sb.charAt(i));
            char cc = c.charValue();
            int nc = (int) cc;

            result = result + "" + c;

            if (nc == 92) {
                result = result + "\\";
            }
        }
        return result;
    }

    // Launch DR-DEVICE.
    private void runBat(String projectName) {
        System.out.println("Launching DR-DEVICE.");

        // Execute "run.bat"
        Runtime rt = Runtime.getRuntime();
        String commandArray[] = {
            //"cmd /c cd C:\\Progra~1\\DR-Device\\",
            //"cmd /c start /B C:\\Progra~1\\DR-Device\\run.bat"
            "cmd /c start /B run.bat"
        };
        try {
            rt.exec("cmd /c start C:\\Progra~1\\DR-Device\\projects\\" + projectName + "\\run-" + projectName + ".bat");
        } catch (IOException ioe) {
            System.out.println("Could not execute run.bat");
            System.out.println(ioe.getMessage());
            ioe.printStackTrace();
            return;
        }
    }

    // Returns the contents of an RDF file as a string.
    private String rdfFileContent() {

        System.out.println("Retrieving results.");

        // Test if the file exists and is in its final state.
        File file = new File("C:\\Program Files\\DR-Device\\projects/" + projectName + "/" + export);

        // Check that the file is finalized
        while (!file.exists()) {
        }
        while (file.length() == 0) {
        }
        boolean test = true;
        while (test) {
            long size1, size2;
            size1 = file.length();
            try {
                Thread.sleep(500);
            } catch (InterruptedException ie) {
            }
            size2 = file.length();
            if (size1 == size2) {
                test = false;
            }
        }

        Integer length = (int) file.length();
        String str = "";
        try {
            byte[] data = new byte[length];
            FileInputStream f = null;
            DataInputStream df = null;
            f = new FileInputStream(file);
            df = new DataInputStream(f);
            df.read(data);
            for (int i = 0; i < data.length; i++) {
                str = str + (char) data[i];
            }
            df.close();
            f.close();
        } catch (Exception e) {
            System.out.println("There is no such file...");
            System.out.println(e.getMessage());
        }
        return str;
    }

    /**
     * Is used to remember the order of variables so that the message returned
     * to the OA does not contain randomly ordered data.
     *
     * @param message The RuleML message
     *
     * @return An array containing the order of the variables.
     */
    public static String[] getVariableOrder(String message) {
        Vector<String> variables = new Vector<String>();
        String[] variableList;
        //Break string up
        StringTokenizer st = new StringTokenizer(message, "<");

        String temp = ""; //Temporary storage
        String tempVar = ""; //Temporary variable storage

        //While information remains
        while (st.hasMoreTokens()) {
            temp = st.nextToken();
            //If it is a variable
            if (temp.startsWith("Var>")) {
                tempVar = "";
                //Get the name of the variable
                for (int i = 4; i < temp.length(); i++) {
                    if (temp.charAt(i) == '<') {
                        break;
                    } else {
                        tempVar = tempVar + temp.charAt(i);
                    }
                }
                variables.addElement(tempVar); //Store the variable name
            }
        }
        //Convert the vector to an array
        variableList = new String[variables.size()];
        for (int i = 0; i < variables.size(); i++) {
            variableList[i] = variables.elementAt(i);
        }
        return variableList;
    }
}